#include <stdio.h>
#include <memory.h>
#include <inttypes.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <libusb-1.0/libusb.h>

#include "../lib/optparse.h"
#include "utils.h"
#include "skiller.h"
#include "usb.h"

#define USBMON_IOC_MAGIC 0x92
#define USBMON_IOCX_GET	 _IOW(USBMON_IOC_MAGIC, 10, usbmon_get_t)

typedef struct usbmon_packet {
	uint64_t id;			    /*  0: URB ID - from submission to callback */
	uint8_t  type;	            /*  8: Same as text; extensible. */
	uint8_t  xfer_type;         /*    ISO (0), Intr, Control, Bulk (3) */
	uint8_t  epnum;	            /*    Endpoint number and transfer direction */
	uint8_t  devnum;	        /*    Device address */
	uint16_t busnum;		    /* 12: Bus number */
	char flag_setup;	        /* 14: Same as text */
	char flag_data;		        /* 15: Same as text; Binary zero is OK. */
	int64_t ts_sec;		        /* 16: gettimeofday */
	uint64_t ts_usec;		    /* 24: gettimeofday */
	int status;		            /* 28: */
	unsigned int length;	    /* 32: Length of data (submitted or actual) */
	unsigned int len_cap;	    /* 36: Delivered length */
	union {			            /* 40: */
		unsigned char setup[8];	/* Only for Control S-type */
		struct iso_rec {		/* Only for ISO */
			int error_count;
			int numdesc;
		} iso;
	} s;
	int interval;		        /* 48: Only for Interrupt and ISO */
	int start_frame;	        /* 52: For ISO */
	unsigned int xfer_flags;    /* 56: copy of URB's transfer_flags */
	unsigned int ndesc;	        /* 60: Actual number of ISO descriptors */
} usbmon_packet_t;				/* 64 total length */

typedef struct usbmon_get {
	usbmon_packet_t *packet;
	void            *data;
	size_t          alloc;
} usbmon_get_t;

const char types[4] = "SICB";

uint8_t buffer[4096];
usbmon_packet_t packet;

int main(int argc, char** argv) 
{
    libusb_device **list;

    libusb_init(NULL);
    size_t count = libusb_get_device_list(NULL, &list);
    libusb_device *skiller = skillerctl_find_device(list, count);

    if (skiller != NULL) {
        uint8_t bus = libusb_get_bus_number(skiller);
        uint8_t dev = libusb_get_device_address(skiller);

        printf("Keyboard found on bus %03d, device: %d\n", bus, dev);

        char* device = "/dev/usbmon0";

        usbmon_get_t get = {
            .packet = &packet,
            .data   = &buffer,
            .alloc  = sizeof buffer
        };

        int fd = open(device, O_RDONLY);
        
        if (fd < 0) {
            printf("Cannot open device %s, maybe usbmon module is not loaded corectly?\n", device);
            return 1;
        }

        while (1) {
            if (ioctl(fd, USBMON_IOCX_GET, &get) < 0) {
                break;
            }

            if (packet.devnum != dev || packet.busnum != bus) {
                continue;
            }

            printf("%s dev %d [%c] len: %d \n", packet.type == 'S' ? "->" : "<-", packet.devnum, types[packet.xfer_type], packet.length);

            skillerctl_packet_t ctl;
            if (skillerctl_unpack(&ctl, buffer, packet.length)) {
                printf("Command: 0x%02X, checksum: 0x%04X, payload length: %d, addr: 0x%04X \n", ctl.command, ctl.checksum, ctl.length, ctl.address);
                hexdump(ctl.payload, ctl.length);
            } else if (packet.length > 0) {
                printf("Unknown packet kind! \n");
                hexdump(buffer, packet.length);
            }
        }
    }


    libusb_free_device_list(list, 1);
    libusb_exit(NULL);
}
